/*
 This is a basic driver for the SSD1320 OLED driver IC. It can initialize the display
 and do simple commands like setPixels, drawCharacters, and other simple commands.

 We can't use SPI library out of the box because the display
 requires 9-bit data in 3-wire SPI mode.
 So we bit-bang the first bit then use SPI hardware to send remaining 8 bits
*/

#include "SSD1320_OLED.h"

// Add header of the fonts here.  Remove as many as possible to conserve FLASH memory.
#include "util/font5x7.h"
#include "util/font8x16.h"
#include "util/7segment.h"
#include "util/fontlargenumber.h"

// Change the total fonts included
#define TOTALFONTS    4

// Add the font name as declared in the header file. Remove as many as possible to conserve FLASH memory.
const unsigned char *SSD1320::fontsPointer[] = {
  font5x7,
  font8x16,
  sevensegment,
  fontlargenumber
};

// Definition of D/C# command bits
#define SPI3_COMMAND LOW   // Command bit is LOW
#define SPI3_DATA    HIGH  // Data bit is HIGH

enum {
  OLED_INTERFACE_SPI3, // 3-wire SPI interface
  OLED_INTERFACE_I2C   // I2C interface
};


/** \brief Grayscale Flexible OLED screen buffer.
  Page buffer 80 x 32 = 2,560 bytes are needed for full 4-bit grayscale. We don't have that.
  So the buffer is 1 bit per pixel or 640 pixels

  TODO - Allow for turning on/off different buffers for different targets

  Page buffer is required because in SPI mode the host cannot read the SSD1320's GDRAM
  of the controller.  This page buffer serves as a scratch RAM for graphical functions.
  All drawing function will first be drawn on this page buffer, only upon calling
  display() function will transfer the page buffer to the actual LCD controller's memory.
*/
static uint8_t screenMemory [] = {
  //LCD Memory organized in 20 bytes (160 columns) and 32 rows = 640 bytes

  //SparkFun Electronics LOGO

  // ROW0, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW1, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW2, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW3, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW4, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW5, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW6, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW7, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW8, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW9, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW10, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW11, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW12, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW13, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW14, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW15, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW16, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW17, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW18, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,

  // ROW19, BYTE0 to BYTE20
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xE0, 0xF8, 0xFC, 0xFE, 0xFF, 0xFF, 0xFF, 0xFF,
  0xFF, 0xFF, 0xFF, 0x0F,
};

SSD1320::SSD1320(uint8_t csPin, uint8_t rstPin, uint8_t sclkPin, uint8_t sdoutPin, SPIClass *spiInterface) {
  _cs = csPin;
  _rst = rstPin;
  _sclk = sclkPin;
  _sd = sdoutPin;

  _interface = OLED_INTERFACE_SPI3;
  _spi = spiInterface;
}

void SSD1320::begin() {
  if (_interface == OLED_INTERFACE_SPI3) {
    pinMode(_cs, OUTPUT);
    digitalWrite(_cs, HIGH); // CS = OUTPUT, init HIGH
    pinMode(_rst, OUTPUT);
    digitalWrite(_rst, LOW); // RST = OUTPUT, init LOW
    pinMode(_sclk, OUTPUT);
    digitalWrite(_sclk, LOW);// SCLK = OUTPUT, init LOW
    pinMode(_sd, OUTPUT);
    digitalWrite(_sd, LOW);  // SDIN = OUTPUT, init LOW

  } else if (_interface == OLED_INTERFACE_I2C) {
    //! TODO
  }

  setFontType(0);
  setColor(WHITE);
  setDrawMode(NORM);
  setCursor(0, 0);

  //memset(displayBuffer, 0xAA, SSD1320_HEIGHT * SSD1320_WIDTH / 4);

  powerUp();
}

///////////////////////
// Private Functions //
///////////////////////

// Execute power up sequence as diagramed on page 11 of OLED datasheet
void SSD1320::powerUp() {
  digitalWrite(_rst, LOW); // Start with display off
  delay(1); // Wait for power to stabilize

  digitalWrite(_rst, HIGH); // Set reset line high
  delayMicroseconds(3);

  command(SSD1320_CMD_OFF);       // 0xAE - Display off

  command(SSD1320_CMD_DIV_RATIO); // 0xD5 - Clock divide ratio/osc. freq
  command(0xC2);                     // 0xC2 - osc clock=0xC divide ratio = 0x2

  command(SSD1320_CMD_MULTIPLEX); // 0xA8 - Multiplex ratio
  command(0x1F);                     // 0x1F - 31

  command(SSD1320_CMD_DISPLAY_OFFSET); // 0xD3 - Display offset
  command(0x60);                          // 0x60 - 96

  command(SSD1320_CMD_START_LINE); // 0xA2 - Start line
  command(0x00);                      // 0x00 - Line 0

  command(SSD1320_CMD_SEG_REMAP);  // 0xA0 - Segment re-map

  command(SSD1320_CMD_COM_SCAN_DIR); // 0xC0 - COM Output scan direction

  command(SSD1320_CMD_SET_SEG_CFG); // 0xDA - seg pins hardware config
  command(0x12);                       // 0x12 -

  command(SSD1320_CMD_CONTRAST);    // 0x81 - Contrast control
  command(0x5A);                       // 0x5A - value between 0x00 and 0xFF

  command(SSD1320_CMD_PRECHARGE_T); // 0xD9 - Pre-charge period
  command(0x22);                       // 0x22

  command(SSD1320_CMD_SET_VCOMH);   // 0xDB - VCOMH Deselect level
  command(0x30);                       // 0x30

  command(SSD1320_CMD_IREF_EN);     // 0xAD - Internal IREF Enable
  command(0x10);                       // 0x10

  command(SSD1320_CMD_MEM_ADD_MODE); // 0x20 - Memory addressing mode
  command(0x00);                        // 0x00 - Horizontal

  // disable internal charge pump
  command(SSD1320_CMD_INT_CHG_PUMP_01); // 0x8D - Internal charge pump
  command(0x01);                           // 0x01
  command(SSD1320_CMD_INT_CHG_PUMP_02); // 0xAC - Internal charge pump
  command(0x00);                           // 0x00

  // set entire display on/off
  command(SSD1320_CMD_DISPLAY_ON);      // 0xA4 - Display on

  // set normal/inverse display
  command(SSD1320_CMD_NORMAL_DISPLAY);  // 0xA6 - Normal display (not inverted)

  // display on
  command(SSD1320_CMD_ON);         // 0xAF - Display on

  //Set the row and column limits for this display
  //These commands also set the RAM pointer on the display to 0,0
  setColumnAddress(0);
  setRowAddress(0);
}

/** \brief Send the display a command byte

  Send a command via SPI, I2C or parallel to SSD1320 controller.
  Display is configured for 3 wire SPI. Arduino SPI does not support 9-bit SPI.
  So we bit bang the first D/C# bit and then use hardware SPI for the command
*/
void SSD1320::command(uint8_t cmd) {

  if (_interface == OLED_INTERFACE_SPI3) {
    digitalWrite(_cs, LOW);  // CS LOW

    // Send D/C# bit
    digitalWrite(_sd, SPI3_COMMAND); // Send command bit
    digitalWrite(_sclk, HIGH); // SCLK HIGH - clock in on rising edge
    digitalWrite(_sclk, LOW);  // SCLK LOW

    // Send command byte
    _spi->beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0)); //Start up SPI again
    _spi->transfer(cmd);
    _spi->endTransaction();

    _spi->end(); //We have to stop the SPI hardware before we can bit bang the 9th bit

    digitalWrite(_cs, HIGH); // CS HIGH

  } else if (_interface == OLED_INTERFACE_I2C) {
    //! TODO
  }
}

/** \brief Send the display a data byte

  Send a command via SPI, I2C or parallel to SSD1320 controller.
  Display is configured for 3 wire SPI. Arduino SPI does not support 9-bit SPI.
  So we bit bang the first D/C# bit and then use hardware SPI for the data
*/
void SSD1320::data(uint8_t cmd) {

  if (_interface == OLED_INTERFACE_SPI3) {
    digitalWrite(_cs, LOW);  // CS LOW

    // Send D/C# bit
    digitalWrite(_sd, SPI3_DATA); // Send data bit
    digitalWrite(_sclk, HIGH); // SCLK HIGH - clock in on rising edge
    digitalWrite(_sclk, LOW);  // SCLK LOW

    // Send data byte
    _spi->beginTransaction(SPISettings(8000000, MSBFIRST, SPI_MODE0)); //Start up SPI again
    _spi->transfer(cmd);
    _spi->endTransaction();

    _spi->end(); //We have to stop the SPI hardware before we can bit bang the 9th bit

    digitalWrite(_cs, HIGH); // CS HIGH
  } else if (_interface == OLED_INTERFACE_I2C) {
    //! TODO
  }
}

/** \brief Transfer display memory.
    Bulk move the screen buffer to the SSD1320 controller's memory so that images/graphics
    drawn on the screen buffer will be displayed on the OLED.
*/
void SSD1320::display(void) {

  //Return CGRAM pointer to 0,0
  setColumnAddress(0);
  setRowAddress(0);

  for (uint8_t rows = 0 ; rows < 32 ; rows++)
  {
    for (uint8_t columns = 0 ; columns < 20 ; columns++)
    {
      uint8_t originalByte = screenMemory[(int)rows * 20 + columns];
      for (uint8_t bitNumber = 8 ; bitNumber > 0 ; bitNumber -= 2)
      {
        uint8_t newByte = 0;
        //Because our scatch buffer is too small to contain 4-bit grayscale,
        //we extrapolate 1 bit onto 4 bits so we pull in two bits to make a byte.
        //We look at each bit in the byte and change it to 0 = 0x00 and 1 = 0x0F.
        if ( (originalByte & (1 << (bitNumber - 1))) != 0) newByte |= 0x0F;
        if ( (originalByte & (1 << (bitNumber - 2))) != 0) newByte |= 0xF0;

        data(newByte);
      }
    }
  }
}

/** \brief Override Arduino's Print.
  Arduino's print overridden so that we can use oled.print().
*/
size_t SSD1320::write(uint8_t c) {
  if (c == '\n')
  {
    cursorY += fontHeight;
    cursorX = 0;
  }
  else if (c == '\r')
  {
    // skip
  }
  else
  {
    drawChar(cursorX, cursorY, c, foreColor, drawMode);
    cursorX += fontWidth + 1;
    if ((cursorX > (SSD1320_WIDTH - fontWidth)))
    {
      cursorY += fontHeight;
      cursorX = 0;
    }
  }

  return 1;
}

/** \brief Set font type.
    Set the current font type number, ie changing to different fonts base on the type provided.
*/
uint8_t SSD1320::setFontType(uint8_t type) {
  if ((type >= TOTALFONTS) || (type < 0))
    return false;

  fontType = type;
  fontWidth = pgm_read_byte(fontsPointer[fontType] + 0);
  fontHeight = pgm_read_byte(fontsPointer[fontType] + 1);
  fontStartChar = pgm_read_byte(fontsPointer[fontType] + 2);
  fontTotalChar = pgm_read_byte(fontsPointer[fontType] + 3);
  fontMapWidth = (pgm_read_byte(fontsPointer[fontType] + 4) * 100) + pgm_read_byte(fontsPointer[fontType] + 5); // two bytes values into integer 16
  return true;
}

/** \brief Draw character.
    Draw character c using current color and current draw mode at x,y.
*/
void  SSD1320::drawChar(uint8_t x, uint8_t y, uint8_t c) {
  drawChar(x, y, c, foreColor, drawMode);
}

/** \brief Draw character with color and mode.
    Draw character c using color and draw mode at x,y.
*/
void  SSD1320::drawChar(uint8_t x, uint8_t y, uint8_t c, uint8_t color, uint8_t mode) {
  // TODO - New routine to take font of any height, at the moment limited to font height in multiple of 8 pixels

  uint8_t rowsToDraw, row, tempC;
  uint8_t i, j, temp;
  uint16_t charPerBitmapRow, charColPositionOnBitmap, charRowPositionOnBitmap, charBitmapStartPosition;

  if ((c < fontStartChar) || (c > (fontStartChar + fontTotalChar - 1))) // no bitmap available for the required c
    return;

  tempC = c - fontStartChar; //Turn user's character into a byte number

  // each row (in datasheet is called a page) is 8 bits high, 16 bit high character will have 2 rows to be drawn
  rowsToDraw = fontHeight / 8; // 8 is LCD's page size, see SSD1306 datasheet
  if (rowsToDraw <= 1) rowsToDraw = 1;

  // the following draw function can draw anywhere on the screen, but SLOW pixel by pixel draw
  if (rowsToDraw == 1) {
    for  (i = 0 ; i < fontWidth + 1 ; i++)
    {
      if (i == fontWidth) // this is done in a weird way because for 5x7 font, there is no margin, this code add a margin after col 5
        temp = 0;
      else
        temp = pgm_read_byte(fontsPointer[fontType] + FONTHEADERSIZE + (tempC * fontWidth) + i);

      //0x7F is the first vertical line of the lowercase letter h
      //The fonts are coming in upside down?
      temp = flipByte(temp);

      //Step through this line of the character checking each bit and setting a pixel
      for (j = 0 ; j < 8 ; j++)
      {
        if (temp & 0x01) {
          setPixel(x + i, y + j, color, mode);
        }
        else {
          setPixel(x + i, y + j, !color, mode);
        }

        temp >>= 1;
      }
    }
    return;
  }

  // font height over 8 bit
  // take character "0" ASCII 48 as example
  charPerBitmapRow = fontMapWidth / fontWidth; // 256/8 = 32 char per row
  charColPositionOnBitmap = tempC % charPerBitmapRow; // = 16
  charRowPositionOnBitmap = int(tempC / charPerBitmapRow); // = 1
  charBitmapStartPosition = (charRowPositionOnBitmap * fontMapWidth * (fontHeight / 8)) + (charColPositionOnBitmap * fontWidth) ;

  for (row = 0 ; row < rowsToDraw ; row++) {
    for (i = 0 ; i < fontWidth ; i++) {
      temp = pgm_read_byte(fontsPointer[fontType] + FONTHEADERSIZE + (charBitmapStartPosition + i + (row * fontMapWidth)));

      //The fonts are coming in upside down
      //Additionally, the large font #1 has padding at the (now) bottom that causes problems
      //The fonts really need to be updated
      temp = flipByte(temp);

      for (j = 0 ; j < 8 ; j++) {
        if (temp & 0x01) {
          setPixel(x + i, y + j + ((rowsToDraw - 1 - row) * 8), color, mode);
        }
        else {
          setPixel(x + i, y + j + ((rowsToDraw - 1 - row) * 8), !color, mode);
        }
        temp >>= 1;
      }
    }
  }
}

/** \brief Draw pixel.
  Draw pixel using the current fore color and current draw mode in the screen buffer's x,y position.
*/
void SSD1320::setPixel(uint8_t x, uint8_t y) {
  setPixel(x, y, foreColor, drawMode);
}

/** \brief Draw pixel with color and mode.
  Draw color pixel in the screen buffer's x,y position with NORM or XOR draw mode.
*/
void SSD1320::setPixel(uint8_t x, uint8_t y, uint8_t color, uint8_t mode) {
  if ((x < 0) || (x >= SSD1320_WIDTH) || (y < 0) || (y >= SSD1320_HEIGHT))
    return;

  int byteNumber = y * (SSD1320_WIDTH / 8) + (x / 8);

  if (mode == XOR)
  {
    screenMemory[byteNumber] ^= (1 << (7 - (x % 8)));
  }
  else //mode = NORM
  {
    if (color == WHITE)
    {
      screenMemory[byteNumber] |= (1 << (7 - (x % 8)));
    }
    else
    {
      screenMemory[byteNumber] &= ~(1 << (7 - (x % 8)));
    }
  }
}

/** \brief Clear screen buffer or SSD1306's memory.
    To clear both RAM and local buffer use CLEAR_ALL
    To clear GDRAM inside the LCD controller use CLEAR_DISPLAY
    To clear local buffer use CLEAR_BUFFER
*/
void SSD1320::clearDisplay(uint8_t mode)
{
  if (mode == CLEAR_DISPLAY || mode == CLEAR_ALL) //Clear the RAM on the display
  {
    //Return CGRAM pointer to 0,0
    setColumnAddress(0);
    setRowAddress(0);

    //Display is 160 pixels long and 32 pixels wide
    //Each byte paints two sequential pixels
    //Each 4-bit nibble is the 4-bit grayscale for that pixel
    //There are only 80 columns because each byte has 2 pixels
    for (int rows = 0 ; rows < 32 ; rows++)
      for (int columns = 0 ; columns < 80 ; columns++)
        data(0x00);

    if (mode == CLEAR_ALL) memset(screenMemory, 0, 640); //Clear the local buffer as well
  }
  else //Clear the local buffer
  {
    memset(screenMemory, 0, 640);   // (20 x 32) = 640 bytes
  }
}

/** \brief Set cursor position.
  OLED's cursor position to x,y.
*/
void SSD1320::setCursor(uint8_t x, uint8_t y) {
  cursorX = x;
  cursorY = y;
}

/** \brief Set color.
    Set the current draw's color. Only WHITE and BLACK available.
*/
void SSD1320::setColor(uint8_t color) {
  foreColor = color;
}

/** \brief Set draw mode.
    Set current draw mode with NORM or XOR.
*/
void SSD1320::setDrawMode(uint8_t mode) {
  drawMode = mode;
}

//https://forum.arduino.cc/index.php?topic=117966.0
uint8_t SSD1320::flipByte(uint8_t c)
{
  c = ((c >> 1) & 0x55) | ((c << 1) & 0xAA);
  c = ((c >> 2) & 0x33) | ((c << 2) & 0xCC);
  c = (c >> 4) | (c << 4) ;

  return c;
}

/** \brief Set SSD1320 row address.
    Send page address command and address to the SSD1320 OLED controller.

    This triple byte command specifies row start address and end address
    of the display data RAM. This command also sets the row address
    pointer to row start address.
*/
void SSD1320::setRowAddress(uint8_t address) {
  command(SSD1320_CMD_SET_ROW_ADR); // Set column address
  command(address); //Set start address
  command(SSD1320_HEIGHT - 1); //Set end address: Display has 32 rows of pixels.
  return;
}

/** \brief Set SSD1320 column address.
    Send page address command and address to the SSD1320 OLED controller.

    This triple byte command specifies column start address and end address
    of the display data RAM. This command also sets the column address
    pointer to row start address.
*/
void SSD1320::setColumnAddress(uint8_t address) {
  command(SSD1320_CMD_SET_COL_ADR); // Set column address
  command(address); //Set start address
  command((SSD1320_WIDTH / 2) - 1); //There are 160 pixels but each byte is 2 pixels. We want addresses 0 to 79.
  return;
}

/*
  Return a pointer to the start of the RAM screen buffer for direct access.
*/
uint8_t *SSD1320::getScreenBuffer(void) {
  return screenMemory;
}


