/*
 This is a basic driver for the SSD1320 OLED driver IC. It can initialize the display
 and do simple commands like setPixels, drawCharacters, and other simple commands.

 We can't use SPI library out of the box because the display
 requires 9-bit data in 3-wire SPI mode.
 So we bit-bang the first bit then use SPI hardware to send remaining 8 bits
*/

#include <Arduino.h>
#include <SPI.h>

#define SCLK_PIN_DEFAULT 13
#define SDOUT_PIN_DEFAULT 11
#define CS_PIN_DEFAULT 10
#define RST_PIN_DEFAULT 8

#define BLACK 0
#define WHITE 1

#define SSD1320_HEIGHT 32
#define SSD1320_WIDTH  160
#define FONTHEADERSIZE    6

#define NORM        0
#define XOR         1

#define CLEAR_ALL         0
#define CLEAR_DISPLAY     1
#define CLEAR_BUFFER      2

typedef enum {
  SSD1320_CMD_MEM_ADD_MODE    = 0x20,
  SSD1320_CMD_SET_COL_ADR     = 0x21,
  SSD1320_CMD_SET_ROW_ADR     = 0x22,
  SSD1320_CMD_CONTRAST        = 0x81,
  SSD1320_CMD_INT_CHG_PUMP_01 = 0x8D,
  SSD1320_CMD_SEG_REMAP       = 0xA0,
  SSD1320_CMD_START_LINE      = 0xA2,
  SSD1320_CMD_DISPLAY_ON      = 0xA4,
  SSD1320_CMD_NORMAL_DISPLAY  = 0xA6,
  SSD1320_CMD_MULTIPLEX       = 0xA8,
  SSD1320_CMD_IREF_EN         = 0xAD,
  SSD1320_CMD_INT_CHG_PUMP_02 = 0xAC,
  SSD1320_CMD_OFF             = 0xAE,
  SSD1320_CMD_ON              = 0xAF,
  SSD1320_CMD_COM_SCAN_DIR    = 0xC0,
  SSD1320_CMD_DISPLAY_OFFSET  = 0xD3,
  SSD1320_CMD_DIV_RATIO       = 0xD5,
  SSD1320_CMD_PRECHARGE_T     = 0xD9,
  SSD1320_CMD_SET_SEG_CFG     = 0xDA,
  SSD1320_CMD_SET_VCOMH       = 0xDB
} ssd1320_cmd;

class SSD1320 : public Print {
  public:
    SSD1320(uint8_t csPin = CS_PIN_DEFAULT,
            uint8_t rstPin = RST_PIN_DEFAULT,
            uint8_t sclkPin = SCLK_PIN_DEFAULT,
            uint8_t sdoutPin = SDOUT_PIN_DEFAULT,
            SPIClass *spiInterface = &SPI);
    void begin();
    virtual size_t write(uint8_t);

    // RAW LCD functions
    void command(uint8_t cmd);
    void data(uint8_t d);
    void setColumnAddress(uint8_t address);
    void setRowAddress(uint8_t address);

    // LCD Draw functions
    void display(void);
    void clearDisplay(uint8_t mode = CLEAR_ALL);
    uint8_t setFontType(uint8_t type);
    void drawChar(uint8_t x, uint8_t y, uint8_t c);
    void drawChar(uint8_t x, uint8_t y, uint8_t c, uint8_t color, uint8_t mode);
    void setCursor(uint8_t x, uint8_t y);
    void setPixel(uint8_t x, uint8_t y);
    void setPixel(uint8_t x, uint8_t y, uint8_t color, uint8_t mode);
    void setColor(uint8_t color);
    void setDrawMode(uint8_t mode);
    uint8_t *getScreenBuffer(void);

  private:
    uint8_t _sclk, _sd, _cs, _rst;
    uint8_t _interface;
    SPIClass *_spi;

    void powerUp();
    static const unsigned char *fontsPointer[];

    uint8_t foreColor, drawMode, fontWidth, fontHeight, fontType, fontStartChar, fontTotalChar, cursorX, cursorY;
    uint16_t fontMapWidth;

    uint8_t flipByte(uint8_t thing);
};
